import networkx as nx
import numpy as np
from numba import njit


@njit(cache=True)
def argmatch(x_ref, x):
    """
    Find indices such that |x[indices] - x_ref| = min!

    Parameters
    ----------
    x_ref : np.ndarray
        1D, sorted
    x : np.ndarray
        1D, sorted
    Returns
    -------
    np.ndarray
    """
    size = np.shape(x_ref)[0]
    out = np.zeros(size, dtype=np.int64)
    ref_ind = 0
    ind = 0
    while x_ref[ref_ind] < x[ind]:
        ref_ind += 1

    while ref_ind < size and ind < x.shape[0] - 1:
        if x[ind] <= x_ref[ref_ind] <= x[ind + 1]:
            if np.abs(x[ind] - x_ref[ref_ind]) < np.abs(
                x[ind + 1] - x_ref[ref_ind]
            ):  # smaller is nearer
                out[ref_ind] = ind
            else:  # bigger is nearer
                out[ref_ind] = ind + 1
            ref_ind += 1
        else:
            ind += 1

    while ref_ind < size:
        out[ref_ind] = x.shape[0] - 1
        ref_ind += 1

    return out


def mask_subsequent_duplicates(x: np.ndarray) -> np.ndarray:
    """
    Calculate mask that removes subsequent duplicates.

    For example if x=[1,1,2,2,3,1,3,3] then x[mask]=[1,2,3,1,3].
    If x has more than one dimensions, the duplicates are removed w.r.t. the first axis.
    For example if x=[[1,1],[1,1],[2,2]] then x[mask]=[[1,1],[2,2]].

    Parameters
    ----------
    x : np.ndarray
        1D or 2D array.

    Returns
    -------
    np.ndarray
    """
    if x.ndim == 1:
        mask = x[:-1] != x[1:]
    elif x.ndim == 2:
        mask = np.any(x[:-1] != x[1:], axis=1)
    else:
        raise ValueError("Only 1D and 2D arrays are supported.")
    mask = np.concatenate((np.array([True]), mask))

    return mask


def calculate_neighbor_list(network: nx.Graph):
    """
    Calculate list of neighbors.

    The i-th element of the list is a numpy array containing the
    node-indices of the neighbors of node i.

    Parameters
    ----------
    network : nx.Graph

    Returns
    -------
    List[np.ndarray]
    """
    neighbor_list = []
    for i in network.nodes():
        neighbor_list.append(np.array(list(network.neighbors(i)), dtype=int))
    return neighbor_list



from copy import deepcopy

import numpy as np
from scipy.integrate import solve_ivp

from ..parameters import CNVMParameters


def calc_rre_traj(
    params: CNVMParameters, c_0: np.ndarray, t_max: float, t_eval=None
) -> tuple[np.ndarray, np.ndarray]:
    """
    Solve the RRE given by parameters, starting from c_0, up to time t_max.

    Solves the ODE using scipy's "solve_ivp".

    Parameters
    ----------
    params : CNVMParameters
    c_0 : np.ndarray
        Initial state, shape=(num_opinions,)
    t_max : float
        End time.
    t_eval : np.ndarray, optional
        Time points, at which the solution should be evaluated.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        1. timepoints, shape=(?,).
        2. c, shape=(?, num_opinions).
    """

    def rhs(t, c):
        out = np.zeros_like(c)
        for m in range(params.num_opinions):
            for n in range(params.num_opinions):
                if n == m:
                    continue

                state_change = np.zeros_like(c)
                state_change[m] = -1
                state_change[n] = 1

                prop = c[m] * (
                    params.r_imit * params.prob_imit[m, n] * c[n]
                    + params.r_noise * params.prob_noise[m, n] / params.num_opinions
                )

                out += prop * state_change
        return out

    sol = solve_ivp(rhs, (0, t_max), c_0, rtol=1e-8, atol=1e-8, t_eval=t_eval)
    return sol.t, sol.y.T


def calc_modified_rre_traj(
    params: CNVMParameters,
    c_0: np.ndarray,
    t_max: float,
    alpha: float = 1.0,
    t_eval=None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Solve the RRE with modified parameters, starting from c_0, up to time t_max.

    The parameters are modified by multiplying the imitation rates `r` with the factor `alpha`.
    For instance, if alpha < 1, this effectively slows the dynamics.

    Parameters
    ----------
    params : CNVMParameters
    c_0 : np.ndarray
        Initial state, shape=(num_opinions,)
    t_max : float
        End time.
    alpha : float
        Factor for modification of imitation rates.
    t_eval : np.ndarray, optional
        Time points, at which the solution should be evaluated.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        1. timepoints, shape=(?,).
        2. c, shape=(?, num_opinions).
    """
    modified_params = deepcopy(params)
    modified_params.change_rates(r=alpha * params.r)
    return calc_rre_traj(modified_params, c_0, t_max, t_eval)



import numpy as np
from numba import njit, prange

from ...utils import argmatch
from ..parameters import CNVMParameters


def sample_stochastic_approximation(
    params: CNVMParameters,
    initial_state: np.ndarray,
    max_time: float,
    num_timesteps: int,
    num_samples: int,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Simulate the opinion shares directly.

    Assumes well-mixedness in the sense that the propensity of
    reaction m -> n is given by
    c[m] * (r[m, n] * c[n] + r_tilde[m, n]).

    This is only true for complete networks.
    For other networks the quality of this approximation may vary.

    Parameters
    ----------
    params : CNVMParameters
    initial_state : np.ndarray
        Initial shares c of shape (num_opinions,).
    max_time : float
    num_timesteps : int
        Number of states in returned trajectory, at equidistant times.
    num_samples: int

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        t with shape (num_timesteps,), c with shape (num_timesteps, num_opinions)
    """
    return _sample_many(
        initial_state,
        max_time,
        params.num_agents,
        params.r,
        params.r_tilde,
        num_timesteps,
        num_samples,
    )


@njit(parallel=True, cache=True)
def _sample_many(
    initial_state: np.ndarray,
    t_max: float,
    num_agents: int,
    r: np.ndarray,
    r_tilde: np.ndarray,
    num_timesteps: int,
    num_samples: int,
):
    t_delta = t_max / (5 * num_timesteps)
    t_out = np.linspace(0, t_max, num_timesteps + 1)
    c_out = np.zeros((num_samples, t_out.shape[0], initial_state.shape[0]))

    for i in prange(num_samples):
        t, c = _simulate(
            initial_state,
            t_max,
            t_delta,
            num_agents,
            r,
            r_tilde,
        )
        t = np.array(t)
        c = make_2d(c)

        t_ind = argmatch(t_out, t)
        c_out[i] = c[t_ind, :]

    return t_out, c_out


@njit
def _simulate(
    initial_state: np.ndarray,
    t_max: float,
    t_delta: float,
    num_agents: int,
    r: np.ndarray,
    r_tilde: np.ndarray,
):
    t = 0
    c = initial_state.copy()
    num_opinions = c.shape[0]

    t_list = [0]
    c_list = [initial_state.copy()]
    props = np.zeros((num_opinions, num_opinions))

    t_store = t_delta
    while t < t_max:
        _update_propensities(props, r, r_tilde, c, num_agents)

        sum_props = np.sum(props)
        t += np.random.exponential(1 / sum_props)

        cum_sum = np.cumsum(props / sum_props)
        reaction = np.searchsorted(cum_sum, np.random.random(), side="right")
        m, n = reaction // num_opinions, reaction % num_opinions
        c[m] -= 1 / num_agents
        c[n] += 1 / num_agents

        if t >= t_store:
            t_store += t_delta
            t_list.append(t)
            c_list.append(c.copy())

    return t_list, c_list


@njit
def _update_propensities(
    props: np.ndarray,
    r: np.ndarray,
    r_tilde: np.ndarray,
    c: np.ndarray,
    num_agents: int,
):
    for m in range(props.shape[0]):
        for n in range(props.shape[1]):
            if m == n:
                continue
            props[m, n] = c[m] * (r[m, n] * c[n] + r_tilde[m, n])
    props *= num_agents


@njit
def make_2d(arraylist):
    n = len(arraylist)
    k = arraylist[0].shape[0]
    a2d = np.zeros((n, k))
    for i in range(n):
        a2d[i] = arraylist[i]
    return a2d



from __future__ import annotations

from typing import Optional, Union

import networkx as nx
import numpy as np
from numpy.typing import NDArray

from ..network_generator import NetworkGenerator


class CNVMParameters:
    """
    Container for the parameters of the Continuous-time Noisy Voter Model (CNVM).

    A node i transitions from its current opinion m to a different opinion n at rate
    r[m,n] * d(i,n) / (d(i)^alpha) + r_tilde[m,n],
    where d(i,n) is the count of opinion n in the neighborhood of agent i, and d(i) the degree of node i.

    Either a network has to specified, or a NetworkGenerator,
    or num_agents, in which case a complete network is used.
    If multiple are given, NetworkGenerator overrules network, and network overrules num_agents.

    The rate parameters r and r_tilde can be given as numpy arrays of shape (num_opinions, num_opinions),
    or as floats, in which case all rates are set to this value.

    (Internally, the CNVM uses an equivalent set of rate parameters: r_imit, r_noise, prob_imit and prob_noise.
    A node i transitions from its current opinion m to a different opinion n at rate
    r_imit * d(i,n) / (d(i)^alpha) * prob_imit[m, n] + r_noise * (1/num_opinions) * prob_noise[m, n].
    These parameters can be provided instead of the usual r and r_tilde.)
    """

    def __init__(
        self,
        num_opinions: int,
        num_agents: Optional[int] = None,
        network: Optional[nx.Graph] = None,
        network_generator: Optional[NetworkGenerator] = None,
        alpha: float = 1.0,
        # rate parameters in style 1
        r: Union[float, NDArray, None] = None,
        r_tilde: Union[float, NDArray, None] = None,
        # rate parameters in style 2
        r_imit: Optional[float] = None,
        r_noise: Optional[float] = None,
        prob_imit: Union[float, NDArray] = 1,
        prob_noise: Union[float, NDArray] = 1,
    ):
        self.num_opinions = num_opinions
        self.alpha = alpha

        # network
        self.num_agents, self.network, self.network_generator = _sanitize_network_input(
            num_agents, network, network_generator
        )

        (
            self.r,
            self.r_tilde,
            self.r_imit,
            self.r_noise,
            self.prob_imit,
            self.prob_noise,
        ) = _sanitize_rates_input(
            num_opinions, r, r_tilde, r_imit, r_noise, prob_imit, prob_noise
        )

    def get_network(self) -> nx.Graph:
        """
        If self.network exists, returns the network.
        Else, if a NetworkGenerator was given, returns a network generated by it.
        Else, returns the complete graph.

        Returns
        -------
        nx.Graph
        """
        if self.network is not None:
            return self.network
        elif self.network_generator is not None:
            return self.network_generator()
        else:
            return nx.complete_graph(self.num_agents)

    def set_network(self, network: nx.Graph) -> None:
        """
        Set new network.

        Only works if no NetworkGenerator was given.

        Parameters
        ----------
        network : nx.Graph
        """
        if self.network_generator is not None:
            raise ValueError("Cannot set network when there is a NetworkGenerator.")
        self.network = network
        self.num_agents = len(network.nodes)

    def update_network_by_generator(self) -> None:
        """
        Update self.network via the NetworkGenerator.

        Only works if a NetworkGenerator was given.
        """
        if self.network_generator is None:
            raise ValueError("No NetworkGenerator was given.")
        self.network = self.network_generator()

    def change_rates(
        self,
        r: Union[float, NDArray, None] = None,
        r_tilde: Union[float, NDArray, None] = None,
    ):
        """
        Change one or both rate parameters.

        If only one argument is given, the other rate parameter stays the same.

        Parameters
        ----------
        r : float | np.ndarray, optional
        r_tilde : float | np.ndarray
        """
        r = r if r is not None else self.r
        r_tilde = r_tilde if r_tilde is not None else self.r_tilde
        (
            self.r,
            self.r_tilde,
            self.r_imit,
            self.r_noise,
            self.prob_imit,
            self.prob_noise,
        ) = _sanitize_rates_input(self.num_opinions, r, r_tilde, None, None, 1, 1)


def _sanitize_network_input(
    num_agents: Optional[int],
    network: Optional[nx.Graph],
    network_generator: Optional[NetworkGenerator],
) -> tuple[int, Optional[nx.Graph], Optional[NetworkGenerator]]:
    if network_generator is not None:
        return (network_generator.num_agents, None, network_generator)
    if network is not None:
        return (network.number_of_nodes(), network, None)
    if num_agents is not None:
        return (num_agents, None, None)
    raise ValueError(
        "Either a network or a NetworkGenerator or num_agents has to be specified."
    )


def _sanitize_rates_input(
    num_opinions: int,
    r: Union[float, NDArray, None],
    r_tilde: Union[float, NDArray, None],
    r_imit: Optional[float],
    r_noise: Optional[float],
    prob_imit: Union[float, NDArray],
    prob_noise: Union[float, NDArray],
) -> tuple[NDArray, NDArray, float, float, NDArray, NDArray]:
    if r is not None and r_tilde is not None:  # style 1
        r = _to_matrix(num_opinions, r)
        r_tilde = _to_matrix(num_opinions, r_tilde)
        if np.min(r) < 0 or np.min(r_tilde) < 0:
            raise ValueError("Rates have to be non-negative.")

        (
            r_imit,
            r_noise,
            prob_imit,
            prob_noise,
        ) = convert_rate_to_cnvm(r, r_tilde)

        return (r, r_tilde, r_imit, r_noise, prob_imit, prob_noise)

    if r_imit is not None and r_noise is not None:  # style 2
        prob_imit = _to_matrix(num_opinions, prob_imit)
        prob_noise = _to_matrix(num_opinions, prob_noise)
        if r_imit < 0 or r_noise < 0:
            raise ValueError("Rates have to be non-negative.")
        if np.min(prob_imit) < 0 or np.max(prob_imit) > 1:
            raise ValueError("Probabilities have to be between 0 and 1.")
        if np.min(prob_noise) < 0 or np.max(prob_noise) > 1:
            raise ValueError("Probabilities have to be between 0 and 1.")

        r = r_imit * prob_imit
        r_tilde = r_noise * prob_noise / num_opinions

        return (r, r_tilde, r_imit, r_noise, prob_imit, prob_noise)

    raise ValueError("Rate parameters have to be provided.")


def _to_matrix(num_opinions: int, r: Union[float, NDArray]) -> NDArray:
    """
    If `r` is a already a NDArray, return `r` but with 0 put on the diagonal.
    If `r` is a float, return a `num_opinions` x `num_opinions` matrix with 0 on the diagonal and `r` everywhere else.
    """
    if isinstance(r, (int, float)):
        r = r * np.ones((num_opinions, num_opinions))
    np.fill_diagonal(r, 0)
    return r


def save_params_as_txt_file(filename: str, params: CNVMParameters):
    """
    Save parameters as a readable .txt file.

    Parameters
    ----------
    filename : str
    params : CNVMParameters
    """
    if filename[-4:] != ".txt":
        this_filename = filename + ".txt"
    else:
        this_filename = filename

    with open(this_filename, "w") as f:
        f.write(f"num_agents = {params.num_agents}\n")
        f.write(f"num_opinions = {params.num_opinions}\n")
        f.write(f"alpha = {params.alpha}\n")
        f.write(f"r_imit = {params.r_imit}\n")
        f.write(f"r_noise = {params.r_noise}\n\n")
        f.write(f"prob_imit =\n {params.prob_imit}\n\n")
        f.write(f"prob_noise =\n {params.prob_noise}\n\n")

        if params.network_generator is not None:
            f.write(f"network_generator = {params.network_generator}\n")
        elif params.network is not None:
            f.write(f"network = {params.network}\n")
        else:
            f.write(f"network = fully connected\n")


def convert_rate_to_cnvm(
    r: NDArray, r_tilde: NDArray
) -> tuple[float, float, NDArray, NDArray]:
    """
    Convert the rates r and r_tilde to the parameters used in the CNVM, i.e., r_imit, r_noise, prob_imit, prob_noise.

    The rates r and r_tilde are defined as:
    An agent i transitions from his current opinion m to a different opinion n at rate
    r[m, n] * d(i,n) / (d(i)^alpha) + r_tilde[m, n]
    where d(i,n) is the count of opinion n in the neighborhood of agent i, and d(i) the degree of node i.

    Parameters
    ----------
    r : np.ndarray
        shape = (num_opinions, num_opinions)
    r_tilde : np.ndarray
        shape = (num_opinions, num_opinions)

    Returns
    -------
    tuple[float, float, np.ndarray, np.ndarray]
        r_imit, r_noise, prob_imit, prob_noise
    """
    num_opinions = r.shape[0]

    # r[m,n] = r_imit * prob_imit[m,n]
    this_r = np.copy(r)
    np.fill_diagonal(this_r, 0)
    r_imit = np.max(this_r)
    if r_imit > 0:
        prob_imit = this_r / r_imit
    else:
        prob_imit = np.zeros((num_opinions, num_opinions))

    # r_tilde[m, n] = r_noise / num_opinions * prob_noise[m, n]
    this_r_tilde = np.copy(r_tilde)
    np.fill_diagonal(this_r_tilde, 0)
    r_noise = np.max(this_r_tilde) * num_opinions
    if r_noise > 0:
        prob_noise = this_r_tilde * num_opinions / r_noise
    else:
        prob_noise = np.zeros((num_opinions, num_opinions))

    return r_imit, r_noise, prob_imit, prob_noise


def convert_rate_from_cnvm(params: CNVMParameters) -> tuple[NDArray, NDArray]:
    """
    Convert the rates used in the CNVM to r and r_tilde.

    The rates r and r_tilde are defined as:
    An agent i transitions from his current opinion m to a different opinion n at rate
    r[m, n] * d(i,n) / (d(i)^alpha) + r_tilde[m, n]
    where d(i,n) is the count of opinion n in the neighborhood of agent i, and d(i) the degree of node i.

    Parameters
    ----------
    params : CNVMParameters

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        r, r_tilde
    """
    r = params.r_imit * params.prob_imit
    r_tilde = params.r_noise * params.prob_noise / params.num_opinions
    return r, r_tilde



import numpy as np
from scipy.integrate import solve_ivp

from ..parameters import CNVMParameters


def calc_pair_approximation_traj(
    params: CNVMParameters,
    c_0: float,
    s_0: float,
    d: float,
    t_max: float,
    t_eval=None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    For a CNVM with 2 opinions, calculate the pair approximation.

    The opinions are referred to as "0" and "1".
    The pair approximation is a system of two ODEs.
    The first ODE describes the evolution of the share of nodes with state 1, which is called "c".
    The second ODE describes the evolution of the variable "s",
    which can be interpreted has 0.5 times the share of active edges.
    So in this package, s can be computed as s = 0.5 * sponet.collective_variables.Interfaces.

    Parameters
    ----------
    params : CNVMParameters
    c_0 : float
        Initial state of c.
    s_0 : float
        Initial state of s.
    t_max : float
        End time.
    t_eval : np.ndarray, optional
        Time points, at which the solution should be evaluated.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        1. timepoints, shape=(?,).
        2. c, shape=(?, 2).
    """

    assert params.num_opinions == 2

    def rhs(_, c_pa):
        # assert isinstance(params.r, np.ndarray)
        # assert isinstance(params.r_tilde, np.ndarray)

        c, s = c_pa

        dc = (
            s * (params.r[0, 1] - params.r[1, 0])
            - c * (params.r_tilde[0, 1] + params.r_tilde[1, 0])
            + params.r_tilde[0, 1]
        )

        ds = -2 * (d - 1) / d * s * s / (1 - c) * params.r[0, 1]
        ds += -2 * (d - 1) / d * s * s / c * params.r[1, 0]
        ds += s * (
            (d - 2) / d * (params.r[0, 1] + params.r[1, 0])
            - 2 * params.r_tilde[0, 1]
            - 2 * params.r_tilde[1, 0]
        )
        ds += c * (params.r_tilde[1, 0] - params.r_tilde[0, 1])
        ds += params.r_tilde[0, 1]

        return np.array([dc, ds])

    sol = solve_ivp(
        rhs, (0, t_max), np.array([c_0, s_0]), rtol=1e-8, atol=1e-8, t_eval=t_eval
    )
    return sol.t, sol.y.T



import multiprocessing as mp
import sys
import time
from typing import Optional

import numpy as np

from sponet.multiprocessing import sample_many_runs

from .collective_variables import CollectiveVariable
from .parameters import Parameters


def sample_moments(
    params: Parameters,
    initial_state: np.ndarray,
    t_max: float,
    num_timesteps: int,
    batch_size: int,
    rel_tol: float = 0.01,
    n_jobs: Optional[int] = None,
    collective_variable: Optional[CollectiveVariable] = None,
    seed: Optional[int] = None,
    filename: Optional[str] = None,
    timeout_seconds: Optional[int] = None,
    verbose: bool = False,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, int]:
    """
    Sample in batches and estimate mean and variance.

    Estimates a confidence interval for the mean and
    samples until the size of the confidence interval relative to the mean is less than a tolerance:
    [4 sigma / sqrt(N)] / mu < rel_tol.

    Parameters
    ----------
    params : Parameters
        Either CNVM or CNTM Parameters.
        If a NetworkGenerator is used, a new network will be sampled for every run.
    initial_states : np.ndarray
        Initial state, shape = (num_agents,).
    t_max : float
        End time.
    num_timesteps : int
        Trajectory will be saved at equidistant time points np.linspace(0, t_max, num_timesteps).
    batch_size : int
        Number of samples per batch.
    rel_tol : float, optional
        Stop when the relative size of the 95% confidence interval is smaller than the tolerance.
    n_jobs : int, optional
        If "None", no multiprocessing is applied. If "-1", all available CPUs will be used.
    collective_variable : CollectiveVariable, optional
        If collective variable is specified, the projected trajectory will be returned
        instead of the full trajectory.
    seed : int, optional
        Seed for random number generation.
        If multiprocessing is used, the subprocesses receive the seeds {seed, seed + 1, ...}.
    filename : str, optional
        Save the results afer each batch using numpy.savez.
        The entries are "t", "mean", "variance", "num_samples", "confidence_size", "rel_error", "time_elapsed".
    timeout_seconds : int, optional
        Stop afer the given time has passed.
    verbose : bool, optional
        Whether to print the progress.

    Returns
    -------
    t, mean, var, num_samples : tuple[np.ndarray, np.ndarray, np.ndarray, int]
        t.shape = (num_timesteps,),
        mean.shape = (num_timesteps, num_agents)
        var.shape = (num_timesteps, num_agents)
    """
    num_a = (
        params.num_agents
        if collective_variable is None
        else collective_variable.dimension
    )

    if seed is None:
        seed = np.random.default_rng().integers(2**31)

    if n_jobs is None:
        n_jobs = 1
    elif n_jobs == -1:
        n_jobs = mp.cpu_count()

    if timeout_seconds is None:
        timeout_seconds = sys.maxsize

    sum_x = np.zeros((num_timesteps, num_a))
    sum_xx = np.zeros((num_timesteps, num_a))
    num_samples = 0

    start = time.time()
    while True:
        if verbose:
            print(f"Total number of samples: {num_samples}.")
        t, x = sample_many_runs(
            params,
            np.array([initial_state]),
            t_max,
            num_timesteps,
            batch_size,
            n_jobs=n_jobs,
            collective_variable=collective_variable,
            seed=seed,
        )
        seed += n_jobs  # each process should get a new seed
        num_samples += batch_size
        sum_x += np.sum(x[0], axis=0)
        sum_xx += np.sum(x[0] ** 2, axis=0)

        mean = sum_x / num_samples
        variance = sum_xx / num_samples - mean**2
        confidence_size = 4 * np.sqrt(variance / num_samples)
        rel_error = np.nanmax(confidence_size / mean)
        if filename is not None:
            np.savez_compressed(
                filename,
                t=t,
                mean=mean,
                variance=variance,
                num_samples=num_samples,
                confidence_size=confidence_size,
                rel_error=rel_error,
                time_elapsed=time.time() - start,
            )
        if verbose:
            print(f"Relative error: {rel_error}.\n")
        if rel_error < rel_tol or time.time() - start >= timeout_seconds:
            return t, mean, variance, num_samples
